# frozen_string_literal: true

##
module MIME
  class Types
    # The released version of the mime-types library.
    VERSION = "3.7.0"
  end

  class Type
    # The released version of the mime-types library.
    VERSION = MIME::Types::VERSION
  end
end
